@@ -2,6 +2,7 @@ from ast import literal_eval
 import os
 import sys
 import glob
+import shutil
 import yaml 
 import re
 import time
@@ -11,10 +12,12 @@ import subprocess
 import ruamel.yaml as yaml
 import argparse
 import json
+import base64
 
 import papermill as pm
 import scrapbook as sb
 import nbformat
+from nbconvert import HTMLExporter
 
 from nb2workflow.health import current_health
 from nb2workflow import workflows
@@ -27,7 +30,7 @@ try:
     from nb2workflow import logstash
     logstasher = logstash.LogStasher()
 except Exception as e:
-    logger.warning("unable to setup logstash %s",repr(e))
+    logger.debug("unable to setup logstash %s",repr(e))
 
     logstasher = None
 
@@ -174,8 +177,27 @@ class NotebookAdapter:
     def output_notebook_fn(self):
         return os.path.join(self.tmpdir,os.path.basename(self.notebook_fn.replace(".ipynb","_output.ipynb")))
 
+    def read(self):
+        return nbformat.reads(open(self.notebook_fn).read(), as_version=4)
+
+    def export_html(self, fn=None):
+        if fn is None:
+            fn = "{}_output.html".format(self.name)
+
+        if False:
+            html_exporter = HTMLExporter()
+            html_exporter.template_file = 'basic'
+            (body, resources) = html_exporter.from_notebook_node(self.read())
+            open(fn, "w").write(body)
+        else:
+            logging.info("converting... {}".format(subprocess.check_call(["jupyter", "nbconvert", "--to", "html", self.output_notebook_fn, '--output', os.path.abspath(fn)])))
+
+        logger.info("exported html to %s", fn)
+
+        return fn
+
     def extract_parameters(self):
-        nb=nbformat.reads(open(self.notebook_fn).read(), as_version=4)
+        nb=self.read()
 
         input_parameters = {}
         system_parameters = {}
@@ -243,13 +265,14 @@ class NotebookAdapter:
 
         
 
-    def execute(self, parameters, progress_bar = True, log_output = True):
+    def execute(self, parameters, progress_bar = True, log_output = True, inplace=False):
         t0 = time.time()
         if logstasher is not None:
             logstasher.log(dict(origin="nb2workflow.execute", event="starting", parameters=parameters, workflow_name=notebook_short_name(self.notebook_fn), health=current_health()))
 
 
-        exceptions = self._execute(parameters, progress_bar, log_output)
+        logger.info("starting job")
+        exceptions = self._execute(parameters, progress_bar, log_output, inplace)
 
         tspent = time.time() - t0
         if logstasher is not None:
@@ -263,12 +286,24 @@ class NotebookAdapter:
 
         return exceptions
 
-    def _execute(self, parameters, progress_bar = True, log_output = True):
-        tmpdir = self.new_tmpdir()
+    def _execute(self, parameters, progress_bar = True, log_output = True, inplace=False):
 
-        logger.info("new tmpdir: %s", tmpdir)
+        if not inplace :
+            tmpdir = self.new_tmpdir()
+            logger.info("new tmpdir: %s", tmpdir)
+
+            try:
+                logger.info(subprocess.check_output(["git","clone",os.path.dirname(os.path.realpath(self.notebook_fn)), tmpdir]))
+            except:
+                logger.info("git clone failed, will attempt copytree")
+
+                os.rmdir(tmpdir)
+
+                shutil.copytree(os.path.dirname(os.path.realpath(self.notebook_fn)), tmpdir)
+        else:
+            tmpdir =os.path.dirname(os.path.realpath(self.notebook_fn))
+            logger.info("executing inplace, no tmpdir is input dir: %s", tmpdir)
 
-        logger.info(subprocess.check_output(["git","clone",os.path.dirname(os.path.realpath(self.notebook_fn)), tmpdir]))
         
         self.update_summary(state="started", parameters=parameters)
 
@@ -337,7 +372,7 @@ class NotebookAdapter:
 
     
     def extract_output_declarations(self):
-        nb=nbformat.reads(open(self.notebook_fn).read(), as_version=4)
+        nb=self.read()
 
         outputs = {}
 
@@ -363,17 +398,19 @@ import scrapbook as sb
 import base64
 import json
 import os
+    
+from nb2workflow.nbadapter import denumpyfy
 
 """
         for output in outputs.keys():
             logger.debug("output: %s",output)
             output_gather_content+="""
 try:
-    sb.glue("{output}",{output})
+    sb.glue("{output}",denumpyfy({output}))
 except Exception as e:
     print("failed to glue {output}", {output})
     print("will glue jsonified")
-    sb.glue("{output}",json.dumps({output}))
+    sb.glue("{output}",json.dumps(denumpyfy({output})))
 """.format(output=output)
 
             output_gather_content+="\nisinstance({output},str) and os.path.exists({output}) and sb.glue(\"{output}_content\",base64.b64encode(open({output},'rb').read()).decode())".format(output=output)
@@ -382,7 +419,7 @@ except Exception as e:
         newcell = nbformat.v4.new_code_cell(source=output_gather_content)
         newcell.metadata['tags'] = ['injected-gather-outputs']
 
-        nb=nbformat.reads(open(self.notebook_fn).read(), as_version=4)
+        nb=self.read()
         nb.cells = nb.cells + [newcell] 
 
         pm.iorw.write_ipynb(nb, self.preproc_notebook_fn)
@@ -426,7 +463,72 @@ def nbinspect(nb_source):
     for n, nba in nbas.items():
         logger.info("%s %s", n, pprint.pprint(nba.extract_parameters(), indent=4))
 
-def nbrun(nb_source, inp):
+
+def nbreduce(nb_source, max_size_mb):
+    cellsize_limit = None
+    largest_cellsize = None
+
+
+    while True:
+        current_size_mb = os.path.getsize(nb_source)/1024./1024
+        logging.info('notebook %s size %.4lg Mb', nb_source, current_size_mb)
+
+
+        nb = nbformat.reads(open(nb_source).read(), as_version=4)
+
+        newcells = []
+        outputs_left = 0
+        for i_cell, cell in enumerate(nb.cells):
+            logging.info('try to reduce CELL #%i', i_cell)
+    
+            try:
+                cellsize = len(json.dumps(cell.outputs))
+            except AttributeError:
+                logging.info("cell has no outputs, ignoring")
+                continue
+
+            if largest_cellsize is None or largest_cellsize < cellsize:
+                largest_cellsize = cellsize
+
+            logging.info("cell size %.5lg largest cell size %.5lg", cellsize, largest_cellsize)
+            #logging.info("cell size %.5lg largest cell size %.5lg; metadata %s", cellsize, largest_cellsize, str(cell.metadata))
+
+            if 'injected-gather-outputs' in cell.metadata.get('tags', []):
+                logging.info("is injected-gather-outputs: skipping")
+                continue
+
+            if cellsize_limit is not None and cellsize >= cellsize_limit:
+                #newcell = nbformat.v4.new_code_cell(source="# redacted")
+                #newcells.append(newcell)
+                logging.info('cleaning cell')
+                cell.outputs = []
+                largest_cellsize = None
+
+
+            if cell.outputs != []:
+                logging.info('this cell has viable outputs')
+                outputs_left += 1
+        
+            newcells.append(cell)
+
+        if outputs_left == 0:
+            logging.info('notebook size %.4lg Mb, and not more outputs left, cleaning aborted', current_size_mb)
+            return
+        else:
+            logging.info('notebook size %.4lg Mb, still has %i outputs: cleaning may continue', current_size_mb,  outputs_left)
+
+        nb.cells = newcells 
+        pm.iorw.write_ipynb(nb, nb_source)
+
+        if current_size_mb < max_size_mb:
+            logging.info('notebook size %.4lg Mb is smaller than required %.5lg Mb, only cleaning gathering', current_size_mb, max_size_mb)
+            return
+        else:
+            logging.info('notebook size %.4lg Mb is larger than required %.5lg Mb, setting cell size limit to the largest cell %.5lg', current_size_mb, max_size_mb, cellsize_limit)
+            cellsize_limit = largest_cellsize
+
+
+def nbrun(nb_source, inp, inplace=False):
 
     nbas = find_notebooks(nb_source)
 
@@ -444,22 +546,98 @@ def nbrun(nb_source, inp):
 
     logging.info("found parameters %s", repr(pars))
 
-    exceptions = nba.execute(pars)
+    exceptions = nba.execute(pars, inplace=inplace)
 
     if len(exceptions) == 0:
         logging.info("execution SUCCESSFUL!")
     else:
         logging.error("FAILED: %s", exceptions)
 
+        with open("{}_exceptions.json".format(nba.name), "w") as f:
+            json.dump(list(map(workflows.serialize_workflow_exception, exceptions)), f)
+
+        fn = nba.export_html()
+        open("{}_output.ipynb".format(nba.name), "wb").write(open(nba.output_notebook_fn, "rb").read())
+
+        raise Exception("FAILED to execute {}: {}".format(nba.name, exceptions))
+
     r={}
     for k,v in nba.extract_output().items():
         r[k.strip()]=repr(v)
 
+    with open("{}_output.json".format(nba.name), "w") as f:
+        json.dump(r, f)
+    
     with open("cwl.output.json", "w") as f:
         json.dump(r, f)
 
+    nbdata = open(nba.output_notebook_fn, "rb").read()
+    nbfn = "{}_output.ipynb".format(nba.name)
+    open(nbfn, "wb").write(nbdata)
+        
+
+    #r['output_notebook'] = nbfn
+    #r['output_notebook_content'] = base64.b64encode(nbdata).decode()
+    
+    htmlfn = "{}_output.html".format(nba.name)
+    nba.export_html(htmlfn)
+    
+    r['output_notebook_html'] = htmlfn
+    r['output_notebook_html_content'] = base64.b64encode(open(htmlfn, "rb").read()).decode()
+
     return r
 
+
+def traverse_structure(structure, modifier):
+    if isinstance(structure, dict):
+        return { k:traverse_structure(v, modifier) for k,v in structure.items() }
+
+    if isinstance(structure, list):
+        return [ traverse_structure(v, modifier) for v in structure ]
+
+    return modifier(structure)
+
+
+
+def denumpyfy(data):
+    import numpy as np
+
+    def numpy_adapter(d):
+        if isinstance(d, np.bool_):
+            return bool(d)
+
+        if isinstance(d, np.ndarray):
+            return d.tolist()
+        
+        if isinstance(d, np.float16) or \
+           isinstance(d, np.float32) or \
+           isinstance(d, np.float64):
+            return float(d)
+
+        if isinstance(d, np.int16) or \
+           isinstance(d, np.int32) or \
+           isinstance(d, np.int64):
+            return int(d)
+
+        return d
+
+    return traverse_structure(
+                data,
+                numpy_adapter,
+           )
+
+def main_reduce():
+    parser = argparse.ArgumentParser(description='Reduce notebook size') 
+    parser.add_argument('notebook', metavar='notebook', type=str)
+    parser.add_argument('maxsizeMb', metavar='max_size_mb', type=float)
+    parser.add_argument('--debug', action="store_true")
+    
+    args = parser.parse_args()
+
+    setup_logging(args.debug)
+
+    nbreduce(args.notebook, args.maxsizeMb)
+
 def main_inspect():
     parser = argparse.ArgumentParser(description='Inspect some notebooks') # run locally, remotely, semantically
     parser.add_argument('notebook', metavar='notebook', type=str)
@@ -491,6 +669,7 @@ def main():
     parser = argparse.ArgumentParser(description='Run some notebooks') # run locally, remotely, semantically
     parser.add_argument('notebook', metavar='notebook', type=str)
     parser.add_argument('--debug', action="store_true")
+    parser.add_argument('--inplace', action="store_true")
     
     parser.add_argument('inputs', nargs=argparse.REMAINDER)
 
@@ -505,7 +684,7 @@ def main():
     setup_logging(args.debug)
 
 
-    nbrun(args.notebook, inputs)
+    nbrun(args.notebook, inputs, inplace=args.inplace)
 
 
 if __name__ == "__main__":
